#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>

/******************** constants ********************/

#define ME_RANGE 32
#define STRIDE (2*ME_RANGE)
#define ME_SIZE (STRIDE*STRIDE)
#define ALG_SIZE 40
#define MAX_ALG_SIZE 142
#define POPULATION 1000
#define THREADS 2

// tuned to fit a working set in Core2's L1 cache. I have no idea about optimal values for other cpus
#define CHUNK_DATA 100
#define CHUNK_CODE 4

#define assert2(cond, ...) if(!(cond)) { printf(__VA_ARGS__); exit(1); }
#define SWAP(x, y) do {typeof(x) t = x; x = y; y = t;} while(0)
#define MIN(x, y) ((x)<(y)?(x):(y))
static inline int clip(int x, int min, int max)
{
    return x < min ? min : x > max ? max : x;
}

#define RESET 1
#define QRESET 2
typedef struct {
    int8_t x, y, reset;
} op_t;

typedef struct {
    int size;
    uint32_t score;
    op_t ops[ALG_SIZE];
} alg_t;

typedef struct {
    int size;
    uint32_t score;
    op_t ops[MAX_ALG_SIZE];
} alg_big_t;

/******************** reference algos ********************/

alg_t nosearch = {0,0};

alg_t dia = {
    40, 0, {
    {-1,0,  RESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
    {-1,0, QRESET}, {1,0}, {0,-1}, {0,1},
}};

alg_big_t hex = {
    44, 0, {
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-1,-1,RESET}, {0,-1}, {1,-1}, {-1,0}, {1,0}, {-1,1}, {0,1}, {1,1},
}};

alg_big_t umh = {
    40+64+38, 0, {
    {-1, 0, RESET}, { 1, 0}, { 0,-1}, { 0, 1},
    {-1,-1}, { 1,-1}, {-1, 1}, { 1, 1},
    {-2, 0}, { 2, 0}, { 0,-2}, { 0, 2},
    {-1,-2}, { 1,-2}, {-2,-1}, { 2,-1},
    {-1, 2}, { 1, 2}, {-2, 1}, { 2, 1},
    {-3, 0}, { 3, 0}, { 0,-3}, { 0, 3},
    { 0,-5}, { 0, 5}, {-5, 0}, { 5, 0}, 
    { 0,-7}, { 0, 7}, {-7, 0}, { 7, 0},
    {-9, 0}, { 9, 0}, {-11,0}, {11, 0},
    {-13,0}, {13, 0}, {-15,0}, {15, 0},
    {-4,2, RESET}, {-4,1}, {-4,0}, {-4,-1}, {-4,-2},
    { 4,2}, { 4,1}, { 4,0}, { 4,-1}, { 4,-2},
    { 2,3}, { 0,4}, {-2,3}, {-2,-3}, { 0,-4}, { 2,-3},
    {-8,4}, {-8,2}, {-8,0}, {-8,-2}, {-8,-4},
    { 8,4}, { 8,2}, { 8,0}, { 8,-2}, { 8,-4},
    { 4,6}, { 0,8}, {-4,6}, {-4,-6}, { 0,-8}, { 4,-6},
    {-12,6}, {-12,3}, {-12,0}, {-12,-3}, {-12,-6},
    { 12,6}, { 12,3}, { 12,0}, { 12,-3}, { 12,-6},
    {  6,9}, { 0,12}, { -6,9}, { -6,-9}, { 0,-12}, { 6,-9},
    {-16,8}, {-16,4}, {-16,0}, {-16,-4}, {-16,-8},
    { 16,8}, { 16,4}, { 16,0}, { 16,-4}, { 16,-8},
    { 8,12}, { 0,16}, {-8,12}, {-8,-12}, { 0,-16}, { 8,-12},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-2,0, RESET}, {2,0}, {-1,2}, {1,2}, {-1,-2}, {1,-2},
    {-1,-1,RESET}, {0,-1}, {1,-1}, {-1,0}, {1,0}, {-1,1}, {0,1}, {1,1},
}};

alg_t shikari1 = {
    32, 0, {
    {0,1, RESET}, {0,-3}, {-11,5}, {0,-6}, {1,0}, {-4,0}, {1,5},
    {3,-5, RESET}, {4,0}, {0,-1}, {0,3}, {-6,-13}, {-3,-9}, {-2,7},
    {0,1, RESET}, {2,-1},
    {-3,-2, RESET}, {12,-4}, {4,5}, {-10,0},
    {-1,-5, RESET}, {-1,1}, {1,1}, {10,-7}, {-1,3},
    {1,-1, RESET}, {-1,0}, {0,2}, {-1,-10},
    {0,-1, RESET}, {-4,-10}, {-1,0},
}};

alg_t shikari2 = {
    14, 0, {
    {14,-4, RESET}, {14,5},
    {2,0, RESET}, {2,-11},
    {0,5, RESET}, {0,-5},
    {14,3, RESET},
    {0,1, RESET}, {0,-2}, {0,11},
    {0,5, QRESET},
    {0,-2, RESET}, {0,1},
    {0,1, QRESET},
}};

/******************** eval code ********************/

#if 0
int eval_mb(alg_t *a, uint16_t *pix)
{
    long i;
    int omx, omy, bmx, bmy, bcost;
    omx = omy = bmx = bmy = ME_RANGE;
    bcost = pix[omx+omy*STRIDE];
    for(i=0; i<a->size; i++) {
        if(a->ops[i].reset == QRESET && omx == bmx && omy == bmy)
            break;
        if(a->ops[i].reset) {
            omx = bmx;
            omy = bmy;
        }
        uint32_t mx = omx + a->ops[i].x;
        uint32_t my = omy + a->ops[i].y;
        if((mx | my) < STRIDE) {
            int cost = pix[mx+my*STRIDE];
            if(cost < bcost) {
                bcost = cost;
                bmx = mx;
                bmy = my;
            }
        }
    }
    return bcost;
}

void eval_alg(alg_t *a, uint16_t *data, long data_size)
{
    long i;
    uint32_t score = 0;
    for(i=1; i<data_size; i+=ME_SIZE+1)
        score += eval_mb(a, data+i);
    a->score = score;
}

void eval_algs(alg_t **a, int na, uint16_t *data, long data_size)
{
    long i;
    int j;
    for(j=0; j<na; j++)
        a[j]->score = 0;
    for(i=1; i<data_size; i+=ME_SIZE+1)
        for(j=0; j<na; j++)
            a[j]->score += eval_mb(a[j], data+i);
}
#else

extern uint8_t op_mv[];
extern uint8_t op_reset[];
extern uint8_t op_qreset[];
extern uint8_t op_setup[];
extern int op_mv_size;
extern int op_reset_size;
extern int op_qreset_size;
extern int op_setup_size;

void *compile(alg_t *a)
{
    int i, j, p=0;
    uint8_t *code = malloc(a->size * (op_mv_size + op_qreset_size) + op_setup_size + 2);
    memcpy(code+p, op_setup, op_setup_size);
    p += op_setup_size;
    for(i=0; i<a->size; i++) {
        if(i > 0) {
            if(a->ops[i].reset == RESET) {
                memcpy(code+p, op_reset, op_reset_size);
                p += op_reset_size;
            } else if(a->ops[i].reset == QRESET) {
                memcpy(code+p, op_qreset, op_qreset_size);
                p += op_qreset_size;
            }
        }
        memcpy(code+p, op_mv, op_mv_size);
        for(j=0; j<op_mv_size-3; j++) {
            uint32_t imm = *(uint32_t*)(op_mv+j);
            if(imm == 0x01234567) {
                *(uint32_t*)(code+p+j) = a->ops[i].x + (a->ops[i].y<<8);
                j += 3;
            } else if(imm == 0x89abcdef) {
                *(uint32_t*)(code+p+j) = a->ops[i].x*2 + a->ops[i].y*STRIDE*2;
                j += 3;
            }
        }
        p += op_mv_size;
    }
    memcpy(code+p, "\xf3\xc3", 2); // repz ret
    return code;
}

void eval_alg(alg_t *a, uint16_t *data, long data_size)
{
    long i;
    uint32_t score = 0;
    int (*code)(uint16_t *pix, int mv) = compile(a);
    for(i=1+ME_RANGE*(STRIDE+1); i<data_size; i+=ME_SIZE+1)
        score += code(data+i, ME_RANGE+(ME_RANGE<<8));
    free(code);
    a->score = score;
}

void eval_algs_chunk(uint32_t (**code)(uint16_t*,int), uint32_t *scores, int na, uint16_t *data, long data_size)
{
    long i;
    int j;
    for(i=0; i<data_size; i+=ME_SIZE+1)
        for(j=0; j<na; j++)
            scores[j] += code[j](data+i, ME_RANGE+(ME_RANGE<<8));
}

void eval_algs(alg_t **a, int na, uint16_t *data, long data_size)
{
    long i;
    int j;
    uint32_t (*code[na])(uint16_t *pix, int mv);
    uint32_t scores[na];
    memset(scores, 0, sizeof(scores));
    for(j=0; j<na; j++) {
        code[j] = compile(a[j]);
    }
    for(i=1+ME_RANGE*(STRIDE+1); i<data_size; i+=CHUNK_DATA*(ME_SIZE+1)) {
        int chunk_data = MIN(data_size-i, CHUNK_DATA*(ME_SIZE+1));
        for(j=0; j<na; j+=CHUNK_CODE)
            eval_algs_chunk(code+j, scores+j, MIN(na-j, CHUNK_CODE), data+i, chunk_data);
    }
    for(j=0; j<na; j++) {
        free(code[j]);
        a[j]->score = scores[j];
    }
}
#endif

typedef struct {
    alg_t **a;
    int na;
    uint16_t *data;
    long data_size;
} eval_thread_t;

void eval_algs_thread1(eval_thread_t *o)
{
    eval_algs(o->a, o->na, o->data, o->data_size);
}

void eval_algs_thread(alg_t **a, int na, uint16_t *data, long data_size)
{
    int threads = MIN(na/4, THREADS);
    pthread_t h[threads];
    eval_thread_t o[threads];
    int i;
    for(i=0; i<threads; i++) {
        o[i].a = a + na*i/threads;
        o[i].na = na*(i+1)/threads - na*i/threads;
        o[i].data = data;
        o[i].data_size = data_size;
        pthread_create(h+i, NULL, (void*(*)(void*))&eval_algs_thread1, o+i);
    }
    for(i=0; i<threads; i++) {
        pthread_join(h[i], NULL);
    }
}

void print_alg(alg_t *a)
{
    int i;
    for(i=0; i<a->size; i++) {
        if(i && a->ops[i].reset)
            printf(a->ops[i].reset == QRESET ? "qreset\n" : "reset\n");
        printf("mv: %2d, %2d\n", a->ops[i].x, a->ops[i].y);
    }
    printf("\n");
}

/******************** genetic algorithm ********************/

int8_t rand_mv_small[32] = {5,4,3,3,2,2,2,1,1,1,1,1,1,0,0,0,0,0,0};
int8_t rand_mv_big[256];

void solve_init()
{
    int i;
    for(i=0; i<16; i++)
        rand_mv_small[16+i] = -rand_mv_small[15-i];
    for(i=0; i<128; i++)
        rand_mv_big[128+i] = 16.0*pow(i/128.0,2.0);
    for(i=0; i<128; i++)
        rand_mv_big[127-i] = -rand_mv_big[128+i];
}

void mutate(alg_t *a, alg_t *sib)
{
    int k = rand() & 3;
    if(k == 0)
        k = 1;
    while(k--) {
        int i = rand() % a->size;
        int j;
        switch(rand()&7) {
            case 0:
            case 1:
            case 2:
                a->ops[i].x = clip(a->ops[i].x + rand_mv_small[rand()&31], -31, 31);
                a->ops[i].y = clip(a->ops[i].y + rand_mv_small[rand()&31], -31, 31);
                break;
            case 3:
            case 4:
                a->ops[i].x = rand_mv_big[rand()&255];
                a->ops[i].y = rand_mv_big[rand()&255];
                break;
            case 5:
                a->ops[i].reset = !a->ops[i].reset;
                break;
            case 6:
                j = rand() % a->size;
                SWAP(a->ops[i], a->ops[j]);
                break;
            case 7:
                j = rand() % (a->size - i);
                memcpy(a->ops+i, sib->ops+i, j*sizeof(a->ops[0]));
                break;
        }
    };
}

int cmp_alg(const void *a, const void *b)
{
    return ((alg_t*)a)->score - ((alg_t*)b)->score;
}

void solve(uint16_t *data, long data_size)
{
    alg_t pop[POPULATION];
    alg_t *popp[POPULATION];
    int i, j;
    pop[0] = dia;
    eval_alg(pop+0, data, data_size);
    for(i=0; i<dia.size; i++)
        pop[0].ops[i].reset = !!pop[0].ops[i].reset;
    for(i=0; i<POPULATION; i++) {
        pop[i] = pop[0];
        popp[i] = pop+i;
    }

    int generation;
    for(generation=1;; generation++) {
        alg_t pop2[POPULATION];
        // keep the best 2%
        // another 8% are selected at random but biased towards better members
        // rest are mutated
        for(i=0; i<POPULATION; i++) {
            int r = i;
            if(i > POPULATION/50 || (i > 0 && !(rand()&15)))
                r = 1+pow(sqrt(POPULATION-1)*rand()/(RAND_MAX+1.0), 2.0);
            pop2[i] = pop[r];
            if(i > POPULATION/10) {
                int r2 = 1+pow(sqrt(POPULATION-1)*rand()/(RAND_MAX+1.0), 2.0);
                mutate(pop2+i, pop+r2);
            } else if(generation > 1) {
                // prevent duplicates
                for(j=0; j<i; j++) {
                    if(!memcmp(pop2+j, pop2+i, sizeof(alg_t))) {
                        i--;
                        break;
                    }
                }
            }
        }
        memcpy(pop, pop2, sizeof(pop));
        eval_algs_thread(popp, POPULATION, data, data_size);
        qsort(pop, POPULATION, sizeof(alg_t), &cmp_alg);
        printf("generation: %d score: %u\n", generation, pop[0].score);
        if(!(generation%10))
            print_alg(pop+0);
        fflush(stdout);
    }
}

/******************** main ********************/

void *mmap_file(const char *name, off_t *size)
{
    int fd = open(name, O_RDONLY);
    assert2(fd>=0, "can't open '%s'\n", name);
    struct stat stats;
    stat(name, &stats);
    *size = stats.st_size;
    assert2(*size>0, "empty '%s'\n", name);
    void *map = mmap(0, *size, PROT_READ, MAP_PRIVATE, fd, 0);
    assert2(map, "can't mmap '%s'\n", name);
    return map;
}

int main(int argc, char **argv)
{
    if(argc != 2)
        return 1;
    long i, j;
    off_t data_size;
    uint16_t *data = mmap_file(argv[1], &data_size);
    if(!data_size || data_size % (2*ME_SIZE+2)) {
        fprintf(stderr, "data_size not a multiple of %#x\n", 2*ME_SIZE+2);
        return 1;
    }
    data_size >>= 1;
    for(i=0; i<data_size; i+=ME_SIZE+1) {
        if(data[i] != ME_RANGE) {
            fprintf(stderr, "sync word %ld is %d, should be %d\n", i, data[i], ME_RANGE);
            return 1;
        }
    }

    uint32_t esa_score = 0;
    for(i=1; i<data_size; i+=ME_SIZE+1) {
        int min = INT_MAX;
        for(j=0; j<ME_SIZE; j++)
            if(min > data[i+j])
                min = data[i+j];
        esa_score += min;
    }
    printf("esa %u\n", esa_score);

    const char *names[6] = {"umh", "hex", "dia", "nop", "ds1", "ds2"};
    alg_t *a[6] = {(void*)&umh, (void*)&hex, &dia, &nosearch, &shikari1, &shikari2};
    eval_algs(a, 6, data, data_size);
    for(i=0; i<6; i++)
        printf("%s %u\n", names[i], a[i]->score);

    srand(0);
    solve(data, data_size);

    return 0;
}
